import 'package:angel_container_generator/angel_container_generator.dart';
import 'package:angel_container/angel_container.dart' as angel_container;
import 'package:flutter/widgets.dart' as widget;
import 'package:rxdart/rxdart.dart';
import 'action.dart';

typedef SetState = void Function();
typedef Epic<S, T> = Observable<EpicEndAction<S, T>> Function(
    Observable<T> payload$);

abstract class Module<S> {
  static angel_container.Container _container;

  static angel_container.Container makeContainer() {
    const reflector = GeneratedReflector();
    final container = angel_container.Container(reflector);
    _container = container;
    return container;
  }

  static widget.State<W>
      createAppState<W extends widget.StatefulWidget, M extends Module>(
          widget.State<W> Function(M ayanamiState) builder,
          {bool singleton = true}) {
    final hasSingleton = _container.has(M);
    final M instance = _container.make(M);
    if (!hasSingleton && singleton) {
      _container.registerSingleton(instance, as: M);
    }
    instance._appState = builder(instance);
    instance.state$.add(instance.state);
    return instance._appState;
  }

  S get state;

  final Map<Function, String> actions = {};

  final Subject<Action> action$ = PublishSubject();

  final Subject<S> state$ = BehaviorSubject();

  widget.State _appState;

  widget.State get appState {
    return _appState;
  }

  EpicEndAction<S, T> Function(T payload) createAction<T>(Epic<S, T> method) {
    final actionName = actions[method];
    return (T payload) => EpicEndAction(DispatchSymbol,
        dispatchAction: Action(actionName, payload));
  }

  Observable<Action<T>> ofMethod<T>(Epic<S, T> method) {
    final actionName = actions[method];
    return action$.where((a) => a.type == actionName).cast<Action<T>>();
  }

  Observable<EpicEndAction<S, Null>> setState(SetState setter) {
    return Observable.fromFuture(widget.WidgetsBinding.instance.endOfFrame)
        .doOnData((_) {
      if (_appState.mounted) {
        // ignore: invalid_use_of_protected_member
        _appState.setState(setter);
        state$.add(state);
      }
    }).map((_) => EpicEndAction(SetStateSymbol, state: state));
  }

  Future<Null> dispose() async {
    await action$.close();
    await state$.close();
  }
}

class EpicEndAction<S, T> extends Action<Null> {
  EpicEndAction(String type, {this.dispatchAction, this.state})
      : super(type, null);

  final Action<T> dispatchAction;

  final S state;
}
